#ifndef __creaImageIOTreeHandlerImageAdder_h_INCLUDED__
#define __creaImageIOTreeHandlerImageAdder_h_INCLUDED__

#include <creaImageIOTreeHandler.h>
#include <creaImageIOSynchron.h>
#include <creaImageIOImageReader.h>
//#include <creaImageIOImageWriter.h>

// Signal/slot mechanism for progress events
#include <boost/signal.hpp>
#include <boost/bind.hpp>


namespace creaImageIO
{
	/**
	* \ingroup Model
	*/

  //=======================================================================
  /// Object which can add images files to a TreeHandler. Is able to parse (recursively) a part of a filesystem to look for known images and load their attributes in order to add the images to a Tree (submission via a TreeHandler::AddBranch)

  class TreeHandlerImageAdder
  {
	  
  public:
    //====================================================================
    /// Ctor
    TreeHandlerImageAdder(TreeHandler* tree);
    /// Dtor
    ~TreeHandlerImageAdder();
    /// Sets the TreeHandler
    void SetTreeHandler(TreeHandler* tree) { mTreeHandler = tree;}
	
	/// Sets the synchronizer
	void SetSynchronizer(Synchronizer* s){mSynchronizer=s;}
	/// Sets the synchronizer
	void SetCurrentDatabase(std::string cur){mCurrentDB=cur;}
    //====================================================================

    //====================================================================
    /// Structure which holds progress information
    /// To stop the image adder use SetStop()
    class Progress
    {
    public:
      Progress() { Reset(); }
      ~Progress() {}
      
      void Reset() 
      {
	mStop = false;
	mNumberScannedFiles = 0;
	mNumberScannedDirs = 0;
	mNumberHandledFiles = 0;
	mNumberAddedFiles = 0;
      }

      int GetNumberScannedFiles() const { return mNumberScannedFiles; }
      int GetNumberScannedDirs() const { return mNumberScannedDirs; }
      int GetNumberHandledFiles() const { return mNumberHandledFiles; }
      int GetNumberAddedFiles() const { return mNumberAddedFiles; }

      void IncNumberScannedFiles() { mNumberScannedFiles++; }
      void IncNumberScannedDirs() { mNumberScannedDirs++; }
      void IncNumberHandledFiles() { mNumberHandledFiles++; }
      void IncNumberAddedFiles() { mNumberAddedFiles++; }

      void SetStop() { mStop = true; }
      bool GetStop() const { return mStop; }

    private:
      bool mStop;
      int mNumberScannedFiles;
      int mNumberScannedDirs;
      int mNumberHandledFiles;
      int mNumberAddedFiles;
    };
    //=============================================

    //=============================================
    const Progress& GetProgress() const { return mProgress; }
    //=============================================

    //=============================================
    typedef boost::signal<void (Progress&)>  ProgressSignalType;
    typedef ProgressSignalType::slot_function_type ProgressCallbackType;
    //=============================================

   //==================================================================
    /// Adds the function f to the list of functions to call 
    /// when the addition progresses.
    /// f is of type ProgressCallbackType which is:
    /// void (*ProgressCallbackType)(Progress&)
    /// To pass a member function 'f' of an instance 'c' of a class 'C' 
    /// as callback you have to 'bind' it, i.e. call:
    /// ConnectProgressObserver ( boost::bind( &C::f , c, _1 ) );
    void ConnectProgressObserver(ProgressCallbackType callback);
   //==================================================================

    //====================================================================
    /// Returns if the file can be read or not
    bool IsHandledFile( const std::string& filename);
    /// Adds a list of files to the TreeHandler 
    void AddFiles( const std::vector<std::string>& filename );
    /// (Recursively) adds the files of a directory to the TreeHandler 
    void AddDirectory( const std::string& directory, 
		       bool recurse);
	/// Removes a file from the databases
	void RemoveFile(tree::Node* node);
	/// Removes files from the databases
	void RemoveFiles(const std::vector<tree::Node*>& nodes);
	/// Synchronizes the DB and disk by repeating the operations the user has done and returns a report
	std::string Synchronize(bool repair, bool checkAttributes);
	///Recursively checks if the directory is synchronized and optionally the state of the attributes
	void CheckSyncDirectory(const std::string &dirpath, 
							bool recursive, 
							bool repair,
							bool checkAttributes,
							std::vector<std::string> &i_ignorefiles,
							std::vector<std::string> & attsModified,
							std::vector<std::string> & newfiles);
	///Copies the files indicated in the vector and updates all databases
	void CopyFiles(const std::vector<std::string>& filenames, const std::string directory  );
	
	///Saves as the files indicated in the vector in a specific directory
	void SaveAs(const std::vector<std::string>& filenames, std::vector<vtkImageData *> i_images);
	///Finds the node that matches the specified parameters
	void FindNode(tree::Node* parent, int level, 
		const std::string& searchParam, 
		const std::string& searchVal, 
		tree::Node*& node);

	///Finds the nodes that partially match the searchVal
	void FindNodePartial(tree::Node* parent, int level, const std::string& searchParam, const std::string& searchVal, tree::Node*& node);
 
	///Checks the attributes of the database against the ones in disk
	void CheckAttributes(bool repair, std::string& file, std::vector<std::string>& attsModified);
	///Deletes the drive with the given name (use for all databases except maintenance and timestamp)
	void DeleteDriveFromMainDB(const std::string& drive);
	///Deletes the drive with the given name (use for maintenance and timestamp databases)
	void DeleteDriveFromOtherDB(const std::string& drive);
	///Edits the given field and sets the new parameters
	void EditField(tree::Node* node, const std::string& name, const std::string& key, const std::string& val);
	///Returns the demanded attributes for the given file
	void GetAttributes(const std::vector<std::string>& params, 
	  const std::string& filename, 
	  std::vector<std::string>& results);
  
  
  
 
    //====================================================================

  private:

    /// Adds a single file to the TreeHandler 
    /// **WITHOUT** testing wether it is handled or not 
    /// hence you have to call IsHandledFile before using AddFile!
    void AddFile( const std::string& filename );

    /// Recursive method which does the main job for AddDirectory
    void AddDirectoryRecursor( const std::string& directory, 
			       bool recurse,
			   const std::string &addKey);

    TreeHandler* mTreeHandler;
	Synchronizer* mSynchronizer;
    ImageReader mReader;
	//ImageWriter mWriter;
	std::string mCurrentDB;
    
    Progress mProgress;
    ProgressSignalType mProgressSignal;


  };
  // EO class TreeHandlerImageAdder
  //=======================================================================


 


} // EO namespace creaImageIO

#include <iostream>
inline std::ostream& operator << ( std::ostream& o, 
			    const creaImageIO::TreeHandlerImageAdder::Progress& p)
{
  o << p.GetNumberScannedFiles() << " files - "
    << p.GetNumberScannedDirs() << " dirs - "
    << p.GetNumberHandledFiles() << " handled -"
    << p.GetNumberAddedFiles() << " added";
  return o;
}

// EOF
#endif  

